unit RebuildScript;

interface

uses
  MainInfo;

function GetInstallScript(Info: TMainInfo):AnsiString;
function GetCompressMethodName(Info: TMainInfo):string;
function GetLanguageFile(Info: TMainInfo; i:integer): AnsiString;
procedure AddEmbeddedFiles(Info: TMainInfo);

implementation
uses Windows, SysUtils, Struct, Main, MyTypes, PathFunc, MD5, SHA1;

procedure Print(var Res: TBigString; s: string);
begin
  Res.AppendString(s);
end;

procedure PrintLn(var Res: TBigString; s: string);
begin
  Res.AppendString(s+#13#10);
end;

function VerToStr(Ver: Cardinal; ServicePack: Word): String;
var
  Digits: Integer;
begin
  with TSetupVersionDataVersion(Ver) do begin
    Digits := 2;
    if Minor mod 10 = 0 then begin
      Dec (Digits);
      Minor := Minor div 10;
    end;
    FmtStr (Result, '%d.%.*d', [Major, Digits, Minor]);
    if Build <> 0 then
      Result := Result + Format('.%d', [Build]);
    if ServicePack <> 0 then begin
      Result := Result + ' Service Pack ' + IntToStr(Hi(ServicePack));
      if Lo(ServicePack) <> 0 then
        Result := Result + Format('.%d', [Lo(ServicePack)]);
    end;
  end;
end;

function RegRootToStr(const re:TSetupRegistryEntry):string;
begin
  with re do begin
    if rootkey=HKEY_CLASSES_ROOT then Result:='HKCR'
    else if rootkey=HKEY_CURRENT_USER then Result:='HKCU'
    else if rootkey=HKEY_LOCAL_MACHINE then Result:='HKLM'
    else if rootkey=HKEY_USERS then Result:='HKU'
    else if rootkey=HKEY_CURRENT_CONFIG then Result:='HKCC'
    else Result:='';
    if Result<>'' then
      if ro32bit in Options then Result:=Result+'32'
      else if ro64bit in Options then Result:=Result+'64';
  end;
end;

function DoubleQuotes(s:string):string;
var i:integer;
begin
  for i:=length(s) downto 1 do
    if s[i]='"' then Insert('"',s,i);
  Result:=s;
end;

function UnUnicode(Info: TMainInfo; s: string):string;
var
  i,n:integer;
begin
  i:=1; Result:='';
  while i<=length(s) do begin
    if Info.Ver>=4202 then n:=(byte(s[i+1]) shl 8) or byte(s[i])
    else n:=byte(s[i]);//(byte(s[i]) shl 8) or byte(s[i+1]);
    if (n>=32) and (n<=127) then Result:=Result+char(n)
    else Result:=Result+'<'+{IntToStr(n)} Format('%4.4x', [n])+'>';
    if Info.Ver>=4202 then inc(i,2) else inc(i);
  end;
end;

procedure StrParam(var Res: TBigString; DisplayName, Value:string; UseQuotes: boolean = true);
begin
  if Value<>'' then
  begin
    if UseQuotes then
      Print(Res, DisplayName + ': "' + DoubleQuotes(Value) + '"; ')
    else
      Print(Res, DisplayName + ': ' + Value + '; ');
  end;
end;

procedure IntParam(var Res: TBigString; DisplayName:string; Value:Int64);
begin
  if Value <> 0 then StrParam(Res, DisplayName, IntToStr(Value), True);
end;

procedure StrConst(var Res: TBigString; DisplayName, Value:string; ShowAlways:boolean=false);
begin
  if ShowAlways or (Value<>'') then PrintLn(Res, DisplayName+'='+Value);
end;

procedure IntConst(var Res: TBigString; DisplayName:string; Value:Int64);
begin
  StrConst(Res, DisplayName, IntToStr(Value));
end;

function Hash2Str(Info: TMainInfo; var SetupHeader: TMySetupHeader) : string;
begin
  if (Info.Ver < 5309) then
    Result := MD5DigestToString(SetupHeader.PasswordHash_MD5)
  else
    Result := SHA1DigestToString(SetupHeader.PasswordHash_SHA1);
end;

function SaltToStr(Salt: TSetupSalt) : string;
var
  i: integer;
begin
  Result := '';
  for i := Low(Salt) to High(Salt) do
  begin
    Result := Result + IntToHex(Salt[i], 2);
  end;
  Result := LowerCase(Result);
end;

function ProcArcsToStr(Arcs: TMySetupProcessorArchitectures) : String;
  function ArchStr(a: TSetupProcessorArchitecture):string;
  begin
    case a of
      paX86         : Result:='x86';
      paAMD64       : Result:='x64';
      paIA64        : Result:='ia64';
    else Result:='';
    end;
  end;
var
  pa: TSetupProcessorArchitecture;
  t: String;
begin
  Result := '';
  for pa:=Low(pa) to High(pa) do
    if pa in Arcs then begin
      t := ArchStr(pa);
      if t<>'' then Result := Result + t + ' ';
    end;
  Result := Trim(Result);  
end;

procedure NewLine(var Res: TBigString);
begin
  PrintLn(Res, '');
end;

procedure PrintComment(var Res: TBigString; Text: String);
begin
  PrintLn(Res, '; ' + Text);
end;

procedure PrintSectionHeader(var Res: TBigString; SectionName: String; PrependEmptyLine: boolean = true);
begin
  if (PrependEmptyLine) then NewLine(Res);
  PrintLn(Res, '[' + SectionName + ']');
end;

procedure PrintFlagsParam(var Res: TBigString; FlagsStr: String);
begin
  if (FlagsStr <> '') then
    Print(Res, 'Flags: ' + FlagsStr);
  NewLine(Res);
end;

procedure PrintVersions(var Res: TBigString; MinVersion, OnlyBelowVersion: TMySetupVersionData);
begin
  with MinVersion do
    if (WinVersion>$04000000) or (NTVersion>$04000000) then
      StrParam(Res, 'MinVersion', VerToStr(WinVersion,0) + ',' + VerToStr(NTVersion,NTServicePack), False);
  with OnlyBelowVersion do
    if (WinVersion<>0) or (NTVersion<>0) then
      StrParam(Res, 'OnlyBelowVersion', VerToStr(WinVersion,0) + ',' + VerToStr(NTVersion,NTServicePack), False);
end;

procedure PrintRegistryEntry(var Res: TBigString; Info: TMainInfo; const re:TSetupRegistryEntry);
  function OptStr(Opt: TSetupRegistryOption):string;
  begin
    case Opt of
      roCreateValueIfDoesntExist         : Result:='createvalueifdoesntexist';
      roUninsDeleteValue                 : Result:='uninsdeletevalue';
      roUninsClearValue                  : Result:='uninsclearvalue';
      roUninsDeleteEntireKey             : Result:='uninsdeletekey';
      roUninsDeleteEntireKeyIfEmpty      : Result:='uninsdeletekeyifempty';
      roPreserveStringType               : Result:='preservestringtype';
      roDeleteKey                        : Result:='deletekey';
      roDeleteValue                      : Result:='deletevalue';
      roNoError                          : Result:='noerror';
      roDontCreateKey                    : Result:='dontcreatekey';
    else Result:='';
    end;
  end;
var
  s,t:string;
  i:dword;
  o:TSetupRegistryOption;
  vType: string;
begin
  with re do begin
    StrParam(Res, 'Root', RegRootToStr(re), false);
    StrParam(Res, 'Subkey', Subkey);
    StrParam(Res, 'ValueName', ValueName);
    t:='';
    if Typ<>rtNone then begin
      case Typ of
        rtString: begin vType:='String'; t:=ValueData; end;
        rtExpandString: begin vType:='ExpandSZ'; t:=ValueData; end;
        rtDword: begin vType:='Dword';
          if Info.Ver>=4104 then t:=ValueData else begin
            Move(ValueData[1],i,4); t:=Format('$%x', [i]);
          end;
        end;
        rtQWord: begin vType:='Qword'; t := ValueData; end;
        rtBinary: begin vType:='Binary';
          t:='';
          for i:=1 to length(ValueData) do t:=t+IntToHex(byte(ValueData[i]),2)+' ';
          if length(t)>0 then SetLength(t,length(t)-1);
        end;
        rtMultiString: begin vType:='MultiSZ'; t:=ValueData;
          repeat
            i:=pos(#0,t);
            if i=0 then break;
            t[i]:='{'; insert('break}',s,i+1);
          until false;
        end;
        else vType:='Unknown';
      end;
    end;
    StrParam(Res, 'ValueType', vType, False);
    StrParam(Res, 'ValueData',t);
    StrParam(Res, 'Components',Components, false);
    StrParam(Res, 'Tasks',Tasks, false);
    StrParam(Res, 'Languages',Languages);
    StrParam(Res, 'Check',Check);
    StrParam(Res, 'BeforeInstall',BeforeInstall);
    StrParam(Res, 'AfterInstall',AfterInstall);

    PrintVersions(Res, MinVersion, OnlyBelowVersion);

    s:='';
    for o:=Low(o) to High(o) do if o in Options then begin
      t:=OptStr(o);
      if t<>'' then s:=s+t+' ';
    end;
    PrintFlagsParam(Res, s);
  end;
end;

procedure PrintRunEntry(var Res: TBigString; const re:TSetupRunEntry);
  function OptStr(Opt: TSetupRunOption):string;
  begin
    case Opt of
      roShellExec           : Result:='shellexec';
      roSkipIfDoesntExist   : Result:='skipifdoesntexist';
      roPostInstall         : Result:='postinstall';
      roUnchecked           : Result:='unchecked';
      roSkipIfSilent        : Result:='skipifsilent';
      roSkipIfNotSilent     : Result:='skipifnotsilent';
      roHideWizard          : Result:='hidewizard';
      roRun32Bit            : Result:='32bit';
      roRun64Bit            : Result:='64bit';
      roRunAsOriginalUser   : Result:='';
    else Result:='';
    end;
  end;
var
  s, t: string;
  o:TSetupRunOption;
begin
  with re do begin
    StrParam(Res, 'Filename',Name);
    StrParam(Res, 'Parameters',Parameters);
    StrParam(Res, 'WorkingDir',WorkingDir);
    StrParam(Res, 'RunOnceId',RunOnceId);
    StrParam(Res, 'StatusMsg',StatusMsg);
    StrParam(Res, 'Description',Description);
    StrParam(Res, 'Components',Components, false);
    StrParam(Res, 'Tasks',Tasks, false);
    StrParam(Res, 'Languages',Languages);
    StrParam(Res, 'Check',Check);
    StrParam(Res, 'BeforeInstall',BeforeInstall);
    StrParam(Res, 'AfterInstall',AfterInstall);

    PrintVersions(Res, MinVersion, OnlyBelowVersion);

    for o:=Low(o) to High(o) do if o in Options then begin
      t:=OptStr(o);
      if t<>'' then s:=s+t+' ';
    end;
    case Wait of
      rwNoWait: s := s + 'nowait';
      rwWaitUntilIdle: s := s + 'waituntilidle';
    end;
    PrintFlagsParam(Res, s);
  end;
end;

procedure PrintTaskEntry(var Res: TBigString; const te:TSetupTaskEntry);
begin
  with te do begin
    StrParam(Res, 'Name',Name);
    StrParam(Res, 'Description',Description);
    StrParam(Res, 'GroupDescription',GroupDescription);
    StrParam(Res, 'Components',Components);
    StrParam(Res, 'Languages',Languages);
    StrParam(Res, 'Check',Check);

    PrintVersions(Res, MinVersion, OnlyBelowVersion);

    NewLine(Res);
  end;
end;

procedure PrintComponentEntry(var Res: TBigString; const ce:TSetupComponentEntry);
var
  s:string;
  i:integer;
begin
  with ce do begin
    StrParam(Res, 'Name',Name);
    StrParam(Res, 'Description',Description);
    s:=Types;
    for i:=1 to length(s) do if s[i]=',' then s[i]:=' ';
    StrParam(Res, 'Types',s);
    IntParam(Res, 'ExtraDiskSpaceRequired',Int64(ExtraDiskSpaceRequired));
    StrParam(Res, 'Languages',Languages);
    StrParam(Res, 'Check',Check);

    PrintVersions(Res, MinVersion, OnlyBelowVersion);

    NewLine(Res);
  end;
end;

procedure PrintTypeEntry(var Res: TBigString; const te:TSetupTypeEntry);
begin
  with te do begin
    StrParam(Res, 'Name',Name);
    StrParam(Res, 'Description',Description);
    StrParam(Res, 'Languages',Languages);
    StrParam(Res, 'Check',Check);

    PrintVersions(Res, MinVersion, OnlyBelowVersion);

    NewLine(Res);
  end;
end;

procedure PrintFileEntry(var Res: TBigString; var CurSlice: integer; Info: TMainInfo; const fe:TSetupFileEntry);
  function OptStr(Opt: TSetupFileOption):string;
  begin
    case Opt of
      foConfirmOverwrite         : Result:='confirmoverwrite';
      foUninsNeverUninstall      : Result:='uninsneveruninstall';
      foRestartReplace           : Result:='restartreplace';
      foDeleteAfterInstall       : Result:='deleteafterinstall';
      foRegisterServer           : Result:='regserver';
      foRegisterTypeLib          : Result:='regtypelib';
      foSharedFile               : Result:='sharedfile';
      foCompareTimeStamp         : Result:='comparetimestamp';
      foFontIsntTrueType         : Result:='fontisnttruetype';
      foSkipIfSourceDoesntExist  : Result:='skipifsourcedoesntexist';
      foOverwriteReadOnly        : Result:='overwritereadonly';
      foOverwriteSameVersion     : Result:='';
      foCustomDestName           : Result:='';
      foOnlyIfDestFileExists     : Result:='onlyifdestfileexists';
      foNoRegError               : Result:='noregerror';
      foUninsRestartDelete       : Result:='uninsrestartdelete';
      foOnlyIfDoesntExist        : Result:='onlyifdoesntexist';
      foIgnoreVersion            : Result:='ignoreversion';
      foPromptIfOlder            : Result:='promptifolder';
      foDontCopy                 : Result:='dontcopy';
      foUninsRemoveReadOnly      : Result:='uninsremovereadonly';
      foRecurseSubDirsExternal   : Result:='';
      fo32bit                    : Result:='32bit';
      fo64bit                    : Result:='64bit';
      foExternalSizePreset       : Result:='';
      foSetNTFSCompression       : Result:='setntfscomptression';
      foUnsetNTFSCompression     : Result:='unsetntfscomptression';
      foGacInstall               : Result:='gacinstall';
    else Result:='';
    end;
  end;
var
  s,t:string;
  o:TSetupFileOption;
  Opts:TMySetupFileOptions;
begin
  if (fe.FileType <> ftUserFile) then exit;
  with fe do begin
    if LocationEntry<>-1 then
      with PSetupFileLocationEntry(Info.Entries[seFileLocation][LocationEntry])^ do begin
        if FirstSlice<>LastSlice then
          PrintComment(Res, 'the following file spans ' + GetSliceName(Info, FirstSlice) + ' to ' + GetSliceName(Info, LastSlice))
        else if FirstSlice<>CurSlice then
          PrintComment(Res, 'the following file starts on ' + GetSliceName(Info, FirstSlice));
        CurSlice:=LastSlice;
      end;
    if (LocationEntry=-1) then Print(Res, '; ');
    StrParam(Res, 'Source',SourceFilename);
    StrParam(Res, 'DestDir',RemoveBackslashUnlessRoot(DestDir));
    StrParam(Res, 'DestName',DestName);
    StrParam(Res, 'FontInstall',InstallFontName);
    StrParam(Res, 'Components',Components, false);
    StrParam(Res, 'Tasks',Tasks, false);
    StrParam(Res, 'Languages',Languages);
    StrParam(Res, 'Check',Check);
    StrParam(Res, 'BeforeInstall',BeforeInstall);
    StrParam(Res, 'AfterInstall',AfterInstall);

    PrintVersions(Res, MinVersion, OnlyBelowVersion);

    s:='';
    Opts:=Options;
    if foDeleteAfterInstall in Opts then Exclude(Opts, foUninsNeverUninstall);
    for o:=Low(o) to High(o) do if o in Opts then begin
      t:=OptStr(o);
      if t<>'' then s:=s+t+' ';
    end;
    PrintFlagsParam(Res, s);
  end;
end;

procedure PrintDeleteEntry(var Res: TBigString; const de: TSetupDeleteEntry);
  function TypeStr(Typ: TSetupDeleteType):string;
  begin
    case Typ of
      dfFiles             : Result:='files';
      dfFilesAndOrSubdirs : Result:='filesandordirs';
      dfDirIfEmpty        : Result:='dirifempty';
    end;  
  end;
begin
  with de do begin
    StrParam(Res, 'Type', TypeStr(DeleteType), false);
    StrParam(Res, 'Name',Name);
    StrParam(Res, 'Components',Components, false);
    StrParam(Res, 'Tasks',Tasks, false);
    StrParam(Res, 'Languages',Languages);
    StrParam(Res, 'Check',Check, false);
    StrParam(Res, 'BeforeInstall',BeforeInstall);
    StrParam(Res, 'AfterInstall',AfterInstall);
    NewLine(Res);
  end;
end;

procedure PrintCustomMessageEntry(var Res: TBigString; Info: TMainInfo; const ce:TSetupCustomMessageEntry);
var
  s,v:string;
begin
  with ce do begin
    s := Name;
    if (LangIndex >= 0) then
      s:=PSetupLanguageEntry(Info.Entries[seLanguage][LangIndex])^.Name+'.'+s;
    v := StringReplace(Value, #13#10, '%n', [rfReplaceAll]);
    StrConst(Res, s, v);
  end;
end;

procedure PrintLangOptions(var Res: TBigString; Info: TMainInfo; const le:TSetupLanguageEntry);
begin
  with le do begin
    StrConst(Res, 'LanguageName', UnUnicode(Info, LanguageName));
    StrConst(Res, 'LanguageID', '$'+IntToHex(LanguageID,4));
    IntConst(Res, 'LanguageCodePage', LanguageCodePage);
    StrConst(Res, 'DialogFontName', DialogFontName);
    StrConst(Res, 'TitleFontName', TitleFontName);
    StrConst(Res, 'WelcomeFontName', WelcomeFontName);
    StrConst(Res, 'CopyrightFontName', CopyrightFontName);
    IntConst(Res, 'DialogFontSize', DialogFontSize);
    IntConst(Res, 'TitleFontSize', TitleFontSize);
    IntConst(Res, 'WelcomeFontSize', WelcomeFontSize);
    IntConst(Res, 'CopyrightFontSize', CopyrightFontSize);
    if (RightToLeft) then StrConst(Res, 'RightToLeft', 'yes');
  end;
end;

procedure PrintIconEntry(var Res: TBigString; const ie:TSetupIconEntry);
var
  s:string;
begin
  with ie do begin
    StrParam(Res, 'Name',IconName);
    StrParam(Res, 'Filename',Filename);
    StrParam(Res, 'Parameters',Parameters);
    StrParam(Res, 'WorkingDir',WorkingDir);
    StrParam(Res, 'IconFilename',IconFilename);
    if IconIndex<>0 then StrParam(Res, 'IconIndex', IntToStr(IconIndex), false);
    StrParam(Res, 'Comment',Comment);
    StrParam(Res, 'Components',Components, false);
    StrParam(Res, 'Tasks',Tasks, false);
    StrParam(Res, 'Languages',Languages);
    StrParam(Res, 'Check',Check);
    StrParam(Res, 'BeforeInstall',BeforeInstall);
    StrParam(Res, 'AfterInstall',AfterInstall);

    PrintVersions(Res, MinVersion, OnlyBelowVersion);

    s:='';
    case CloseOnExit of
      icYes: s:=s+'closeonexit ';
      icNo: s:=s+'dontcloseonexit ';
    end;
    case ShowCmd of
      SW_SHOWMAXIMIZED: s:=s+'runmaximized ';
      SW_SHOWMINNOACTIVE: s:=s+'runminimized ';
    end;
    PrintFlagsParam(Res, s);
  end;
end;

procedure PrintDirEntry(var Res: TBigString; const de: TSetupDirEntry);
  function OptStr(Opt: TSetupDirOption):string;
  begin
    case Opt of
      doUninsNeverUninstall      : Result:='uninsneveruninstall';
      doDeleteAfterInstall       : Result:='deleteafterinstall';
      doUninsAlwaysUninstall     : Result:='uninsalwaysuninstall';
      doSetNTFSCompression       : Result:='setntfscomptression';
      doUnsetNTFSCompression     : Result:='unsetntfscomptression';
    else Result:='';
    end;
  end;
var
  s,t:string;
  Opts: TMySetupDirOptions;
  o: TSetupDirOption;
begin
  StrParam(Res, 'Name', de.DirName);
  Opts := de.Options;
  for o:=Low(o) to High(o) do if o in Opts then begin
    t := OptStr(o);
    if t<>'' then s := s + t + ' ';
  end;
  PrintFlagsParam(Res, s);
end;

procedure PrintIniEntry(var Res: TBigString; const ie: TSetupIniEntry);
  function OptStr(Opt: TSetupIniOption):string;
  begin
    case Opt of
      ioCreateKeyIfDoesntExist      : Result:='createkeyifdoesntexist';
      ioUninsDeleteEntry            : Result:='uninsdeleteentry';
      ioUninsDeleteEntireSection    : Result:='uninsdeletesection';
      ioUninsDeleteSectionIfEmpty   : Result:='uninsdeletesectionifempty';
    else Result:='';
    end;
  end;
var
  s,t:string;
  Opts: TMySetupIniOptions;
  o: TSetupIniOption;  
begin
  StrParam(Res, 'FileName', ie.Filename);
  StrParam(Res, 'Section', ie.Section);
  StrParam(Res, 'Key', ie.Entry);
  StrParam(Res, 'String', ie.Value);

  Opts := ie.Options;
  for o:=Low(o) to High(o) do if o in Opts then begin
    t := OptStr(o);
    if t<>'' then s := s + t + ' ';
  end;
  PrintFlagsParam(Res, s);
end;

function GetCompressMethodName(Info: TMainInfo):string;
begin
  case Info.SetupHeader.CompressMethod of
    cmStored: Result:='none';
    cmZip: Result:='zip';
    cmBzip: Result:='bzip2';
    cmLZMA: Result:='lzma';
    cmLZMA2: Result:='lzma2';
  end;
end;

function GetInnoVersionStr(Info: TMainInfo) : string;
begin
  Result := Info.FileVersion;
  if (Info.VerIsUnicode) then
    Result := Result + ' (Unicode)';
end;

function GetBaseFileName(Info: TMainInfo) : string;
var
  name: string;
begin
  name := Info.SetupHeader.BaseFilename;
  if (name = '') then
  begin
    name := ExtractFileName(Info.SetupLdrOriginalFilename);
	// Remove .exe extenstion if have one
	if (Pos('.exe', name) = Length(name) - 3) then
	  name := Copy(name, 1, Length(name) - 4);
  end;
  Result := name;
end;

function MaybeToRtf(const Name, Contents: string):string;
begin
  if copy(Contents, 1, 5)='{\rtf' then
    Result:=ChangeFileExt(Name, '.rtf')
  else
    Result:=Name;
end;

procedure AddFakeRtfOrTxtFile(Info: TMainInfo; const Name, Contents: string);
var
  ContentsA: AnsiString;
  {$IFDEF UNICODE}
  n: integer;
  {$ENDIF}
begin
  {$IFDEF UNICODE}
  n := Length(Contents);
  SetLength(ContentsA, 2*n);
  if n > 0 then
    Move(Contents[1], ContentsA[1], 2*n);
  {$ELSE}
  ContentsA := Contents;
  {$ENDIF}
  AddFakeFile(Info, MaybeToRtf(Name, Contents), ContentsA);
end;

function GetLanguageFile(Info: TMainInfo; i:integer): AnsiString;
var Res: TBigString;
begin
  Res.Init;
  PrintSectionHeader(Res, 'LangOptions', False);
  PrintLangOptions(Res, Info, PSetupLanguageEntry(Info.Entries[seLanguage][i])^);
  Result:= AnsiString(Res.CopyAsString);
  Res.Destroy;
end;

procedure PrintLanguageEntry(var Res: TBigString; Info: TMainInfo; const le:TSetupLanguageEntry);
begin
  with le do begin                                      // these filenames must be mirrored in AddEmbeddedFiles()
    StrParam(Res, 'Name', Name);
    StrParam(Res, 'MessagesFile', 'embedded\'+Name+'.isl');
    if LicenseText<>'' then StrParam(Res, 'LicenseFile', MaybeToRtf('embedded\'+Name+'License.txt', LicenseText));
    if InfoBeforeText<>'' then StrParam(Res, 'InfoBeforeFile', MaybeToRtf('embedded\'+Name+'InfoBefore.txt', InfoBeforeText));
    if InfoAfterText<>'' then StrParam(Res, 'InfoAfterFile', MaybeToRtf('embedded\'+Name+'InfoAfter.txt', InfoAfterText));
  end;
  NewLine(Res);
end;

procedure PrintSetupHeader(var Res: TBigString; Info: TMainInfo);
  function Priv2Str(Priv: TMySetupPrivileges) : string;
  begin
    case Priv of
      prNone        : Result:='none';
      prPowerUser   : Result:='poweruser';
      prLowest      : Result:='lowest';
    else Result:='';
    end;
  end;
  function DisPage2Str(a: TMySetupDisablePage):string;
  begin
    case a of
      dpAuto   : Result:='auto';
      dpNo     : Result:='no';
      dpYes    : Result:='yes';
    else Result:='';
    end;
  end;
begin
  StrConst(Res, ';InnoSetupVersion', GetInnoVersionStr(Info));

  PrintSectionHeader(Res, 'Setup');
  StrConst(Res, 'AppName', Info.SetupHeader.AppName, True);
  StrConst(Res, 'AppVerName', Info.SetupHeader.AppVerName);
  StrConst(Res, 'AppVersion', Info.SetupHeader.AppVersion);
  StrConst(Res, 'AppPublisher', Info.SetupHeader.AppPublisher);
  StrConst(Res, 'AppPublisherURL', Info.SetupHeader.AppPublisherURL);
  StrConst(Res, 'AppSupportPhone', Info.SetupHeader.AppSupportPhone);
  StrConst(Res, 'AppSupportURL', Info.SetupHeader.AppSupportURL);
  StrConst(Res, 'AppUpdatesURL', Info.SetupHeader.AppUpdatesURL);
  StrConst(Res, 'AppMutex', Info.SetupHeader.AppMutex);
  StrConst(Res, 'AppComments', Info.SetupHeader.AppComments);
  StrConst(Res, 'AppModifyPath', Info.SetupHeader.AppModifyPath);
  if (shCreateAppDir in Info.SetupHeader.Options) then
    StrConst(Res, 'DefaultDirName', Info.SetupHeader.DefaultDirName)
  else
    StrConst(Res, 'CreateAppDir', 'no');
  if (Info.SetupHeader.DefaultGroupName <> '(Default)') then
    StrConst(Res, 'DefaultGroupName', Info.SetupHeader.DefaultGroupName);
  StrConst(Res, 'UninstallDisplayIcon', Info.SetupHeader.UninstallDisplayIcon);
  StrConst(Res, 'UninstallDisplayName', Info.SetupHeader.UninstallDisplayName);
  if (Info.SetupHeader.UninstallFilesDir <> '{app}') then
    StrConst(Res, 'UninstallFilesDir', Info.SetupHeader.UninstallFilesDir);
  StrConst(Res, 'OutputBaseFilename',  GetBaseFilename(Info));
  StrConst(Res, 'Compression', GetCompressMethodName(Info));
  if Info.SetupHeader.EncryptionUsed then
    StrConst(Res, '; Encryption', 'yes');
  if (Info.Ver > 4202) and (Info.SetupHeader.EncryptionUsed or (shPassword in Info.SetupHeader.Options)) then
  begin
    StrConst(Res, '; PasswordHash', Hash2Str(Info, Info.SetupHeader));
    StrConst(Res, '; PasswordSalt', SaltToStr(Info.SetupHeader.PasswordSalt));
  end;  
  StrConst(Res, 'ArchitecturesAllowed', ProcArcsToStr(Info.SetupHeader.ArchitecturesAllowed));
  StrConst(Res, 'ArchitecturesInstallIn64BitMode', ProcArcsToStr(Info.SetupHeader.ArchitecturesInstallIn64BitMode));
  if (Info.Ver<5310) and not(shUninstallable in Info.SetupHeader.Options) then
    StrConst(Res, 'Uninstallable', 'no')
  else if not SameText(Info.SetupHeader.Uninstallable, 'yes') then
    StrConst(Res, 'Uninstallable', Info.SetupHeader.Uninstallable);
  if (Info.SetupHeader.PrivilegesRequired <> prAdmin) then
    StrConst(Res, 'PrivilegesRequired', Priv2Str(Info.SetupHeader.PrivilegesRequired));
  if (Int64(Info.SetupHeader.ExtraDiskSpaceRequired) > 0) then
    IntConst(Res, 'ExtraDiskSpaceRequired', Int64(Info.SetupHeader.ExtraDiskSpaceRequired));
  if (Info.SetupHeader.DisableDirPage <> dpNo) then
    StrConst(Res, 'DisableDirPage', DisPage2Str(Info.SetupHeader.DisableDirPage));
  if (Info.SetupHeader.DisableProgramGroupPage <> dpNo) then
    StrConst(Res, 'DisableProgramGroupPage', DisPage2Str(Info.SetupHeader.DisableProgramGroupPage));
  if (shChangesAssociations in Info.SetupHeader.Options) then
    StrConst(Res, 'ChangesAssociations', 'yes');
  if (shAllowNoIcons in Info.SetupHeader.Options) then
    StrConst(Res, 'AllowNoIcons', 'yes');
  // mirror these filenames in AddEmbeddedFiles()
  with Info.SetupHeader do begin
    if LicenseText<>'' then StrConst(Res, 'LicenseFile', MaybeToRtf('embedded\License.txt', string(LicenseText)));
    if InfoBeforeText<>'' then StrConst(Res, 'InfoBeforeFile', MaybeToRtf('embedded\InfoBefore.txt', string(InfoBeforeText)));
    if InfoAfterText<>'' then StrConst(Res, 'InfoAfterFile', MaybeToRtf('embedded\InfoAfter.txt', string(InfoAfterText)));
  end;
  if Info.WizardImage<>'' then StrConst(Res, 'WizardImageFile', 'embedded\WizardImage.bmp');
  if Info.WizardSmallImage<>'' then StrConst(Res, 'WizardSmallImageFile', 'embedded\WizardSmallImage.bmp');
  if Info.TimeStampsInUTC then StrConst(Res, ';TimeStampsInUTC', 'yes');
end;

procedure AddEmbeddedFiles(Info: TMainInfo);
var i:integer;
begin
  with Info.SetupHeader do begin      // mirror these filenames in procedures above
    AddFakeRtfOrTxtFile(Info, 'embedded\License.txt', string(LicenseText));
    AddFakeRtfOrTxtFile(Info, 'embedded\InfoBefore.txt', string(InfoBeforeText));
    AddFakeRtfOrTxtFile(Info, 'embedded\InfoAfter.txt', string(InfoAfterText));
    AddFakeFile(Info, 'embedded\CompiledCode.bin', CompiledCodeText);
  end;
  AddFakeFile(Info, 'embedded\WizardImage.bmp', Info.WizardImage);
  AddFakeFile(Info, 'embedded\WizardSmallImage.bmp', Info.WizardSmallImage);
  AddFakeFile(Info, 'embedded\decompressor.dll', Info.DecompDll);

  for i:=0 to Info.SetupHeader.NumLanguageEntries-1 do
    with PSetupLanguageEntry(Info.Entries[SeLanguage][i])^ do begin
      AddFakeFile(Info, 'embedded\'+Name+'.isl', GetLanguageFile(Info, i));
      if LicenseText<>'' then AddFakeRtfOrTxtFile(Info, 'embedded\'+Name+'License.txt', LicenseText);
      if InfoBeforeText<>'' then AddFakeRtfOrTxtFile(Info, 'embedded\'+Name+'InfoBefore.txt', InfoBeforeText);
      if InfoAfterText<>'' then AddFakeRtfOrTxtFile(Info, 'embedded\'+Name+'InfoAfter.txt', InfoAfterText);
    end;
end;

function GetInstallScript(Info: TMainInfo):AnsiString;
var
  Res:TBigString;
  CurSlice:integer;
  i:integer;
begin
  Res.Init;

  // Add UTF-8 BOM to script start for Unicode versions.
  if (Info.VerIsUnicode) then Print(Res, #$EF#$BB#$BF);

  PrintSetupHeader(Res, Info);

  CurSlice:=0;
  if Info.Entries[seFile].Count>0 then begin
    PrintSectionHeader(Res, 'Files');
    for i:=0 to Info.Entries[seFile].Count-1 do
      PrintFileEntry(Res, CurSlice, Info, PSetupFileEntry(Info.Entries[seFile][i])^);
  end;
  if Info.Entries[seDir].Count>0 then begin
    PrintSectionHeader(Res, 'Dirs');
    for i:=0 to Info.Entries[seDir].Count-1 do
      PrintDirEntry(Res, PSetupDirEntry(Info.Entries[seDir][i])^);
  end;
  if Info.Entries[seRegistry].Count>0 then begin
    PrintSectionHeader(Res, 'Registry');
    for i:=0 to Info.Entries[seRegistry].Count-1 do
      PrintRegistryEntry(Res, Info, PSetupRegistryEntry(Info.Entries[seRegistry][i])^);
  end;
  if Info.Entries[seIni].Count>0 then begin
    PrintSectionHeader(Res, 'INI');
    for i:=0 to Info.Entries[seIni].Count-1 do
      PrintIniEntry(Res, PSetupIniEntry(Info.Entries[seIni][i])^);
  end;
  if Info.Entries[seRun].Count>0 then begin
    PrintSectionHeader(Res, 'Run');
    for i:=0 to Info.Entries[seRun].Count-1 do
      PrintRunEntry(Res, PSetupRunEntry(Info.Entries[seRun][i])^);
  end;
  if Info.Entries[seUninstallRun].Count>0 then begin
    PrintSectionHeader(Res, 'UninstallRun');
    for i:=0 to Info.Entries[seUninstallRun].Count-1 do
      PrintRunEntry(Res, PSetupRunEntry(Info.Entries[seUninstallRun][i])^);
  end;
  if Info.Entries[seIcon].Count>0 then begin
    PrintSectionHeader(Res, 'Icons');
    for i:=0 to Info.Entries[seIcon].Count-1 do
      PrintIconEntry(Res, PSetupIconEntry(Info.Entries[seIcon][i])^);
  end;
  if Info.Entries[seTask].Count>0 then begin
    PrintSectionHeader(Res, 'Tasks');
    for i:=0 to Info.Entries[seTask].Count-1 do
      PrintTaskEntry(Res, PSetupTaskEntry(Info.Entries[seTask][i])^);
  end;
  if Info.Entries[seComponent].Count>0 then begin
    PrintSectionHeader(Res, 'Components');
    for i:=0 to Info.Entries[seComponent].Count-1 do
      PrintComponentEntry(Res, PSetupComponentEntry(Info.Entries[seComponent][i])^);
  end;
  if Info.Entries[seInstallDelete].Count>0 then begin
    PrintSectionHeader(Res, 'InstallDelete');
    for i:=0 to Info.Entries[seInstallDelete].Count-1 do
      PrintDeleteEntry(Res, PSetupDeleteEntry(Info.Entries[seInstallDelete][i])^);
  end;
  if Info.Entries[seUninstallDelete].Count>0 then begin
    PrintSectionHeader(Res, 'UninstallDelete');
    for i:=0 to Info.Entries[seUninstallDelete].Count-1 do
      PrintDeleteEntry(Res, PSetupDeleteEntry(Info.Entries[seUninstallDelete][i])^);
  end;
  if Info.Entries[seType].Count>0 then begin
    PrintSectionHeader(Res, 'Types');
    for i:=0 to Info.Entries[seType].Count-1 do
      PrintTypeEntry(Res, PSetupTypeEntry(Info.Entries[seType][i])^);
  end;
  if Info.Entries[seCustomMessage].Count>0 then begin
    PrintSectionHeader(Res, 'CustomMessages');
    for i:=0 to Info.Entries[seCustomMessage].Count-1 do
      PrintCustomMessageEntry(Res, Info, PSetupCustomMessageEntry(Info.Entries[seCustomMessage][i])^);
  end;
  if (Info.Entries[seLanguage].Count>0) then begin
    PrintSectionHeader(Res, 'Languages');
    PrintComment(Res, 'These files are stubs');
    PrintComment(Res, 'To achieve better results after recompilation, use the real language files');
    for i:=0 to Info.Entries[seLanguage].Count-1 do
      PrintLanguageEntry(Res, Info, PSetupLanguageEntry(Info.Entries[seLanguage][i])^);
  end;
  Result:=AnsiString(Res.CopyAsString);
  Res.Destroy;
end;


end.